#! /usr/bin/perl
#---------------------------------------------------------------------
# $Id: testFFV.pl 4826 2011-11-20 23:28:38Z cjm $
# Copyright 2011 Christopher J. Madsen
#
# Test suite for FFV
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#---------------------------------------------------------------------

use strict;
use warnings;
use Test::More tests => 34;
use File::Temp 'tempdir';
use FindBin    '$Bin';

# Load Test::Differences, if available:
BEGIN {
  if (eval "use Test::Differences; 1") {
    # Not all versions of Test::Differences support changing the style:
    eval { Test::Differences::unified_diff() }
  } else {
    eval '*eq_or_diff = \&is;'; # Just use "is" instead
  }
} # end BEGIN

my $dataDir = "$Bin/testdata";
my $ffv = $ENV{FFV_EXE_PATH};
my $Win32 = ($^O =~ /Win32/);

sub exSuccess       () { 0 }
sub exMissing       () { 1 }
sub exMismatch      () { 2 }
sub exIOError       () { 3 }
sub exInternalError () { 4 }
sub exUsage         () { 5 }

if (not $ffv) {
  # Under Windows, we'll look in the build directories:
  if ($Win32) {
    require File::Spec;

    my ($debug, $release, $shipped) = map { File::Spec->rel2abs($_, $Bin) }
        qw( ../win32/Debug/FFV.exe ../win32/Release/FFV.exe ../FFV.exe );

    if (-e $debug) {
      if (-e $release) {
        # Both $debug & $release exist, use the newest one:
        $ffv = ((-M $debug < -M $release) ? $debug : $release);
      } else {
        $ffv = $debug;          # Only $debug exists
      }
    } elsif (-e $release) {
      $ffv = $release;          # Only $release exists
    } elsif (-e $shipped) {
      $ffv = $shipped;
    }
  } # end if Win32

  $ffv ||= 'ffv';               # Assume it's on the path
} # end

ok(1);
diag("Testing $ffv");

#=====================================================================
# Subroutines:
#---------------------------------------------------------------------
sub file
{
  my ($file) = @_;

  local $/ = undef;

  unless(open(FILE, '<', $file)) {
    warn "Can't open $file: $!";
    return undef;
  }
  my $contents = <FILE>;
  close FILE;

  return $contents;
} # end file

#---------------------------------------------------------------------
# Convert backslashes to forward slashes on Win32:

sub fs
{
  $_[0] =~ s!\\!/!g if $Win32;
  $_[0];
} # end fs

#---------------------------------------------------------------------
sub genfile
{
  my ($file, $seed, $length) = @_;

  system "$Bin/rndfile", $seed, $length, $file and die "Can't create $file";
} # end genfile

#---------------------------------------------------------------------
sub runTest
{
  my ($testName, $args, $status, $expectedOutput) = @_;

  my $output = `$ffv $args 2>&1`;

  local $Test::Builder::Level = $Test::Builder::Level + 1;

  is($? >> 8, $status, "$testName status");

  $output =~ s/ *\r?\n/\n/mg;   # Clean up trailing whitespace
  fs($output);                  # Convert backslashes

  eq_or_diff($output, $expectedOutput, "$testName output");
} # end runTest

#=====================================================================
# Begin testing:
#---------------------------------------------------------------------
# Get version information from configure.ac:

our ($name, $major, $minor);
do "$Bin/getversion.pm";

my $version;

if (defined $major) {
  $version = "$major.$minor";
} else {
  # Fall back to running FFV and asking its version:
  `$ffv --version` =~ /Fast File Validator (.+?) by /
      or die "Can't figure out version of FFV";
  $version = $1;
}

#---------------------------------------------------------------------
# Create directory and files:

my $dir = tempdir(CLEANUP => 1);
#my $dir="/tmp/testFFV"; system "mkdir $dir || rm -rf $dir/*"; # For development
#my $dir="/tmp/testFFV"; use File::Path; rmtree([$dir]); mkdir $dir; # Win32 dev
chdir $dir or die "Failed to change to $dir";

#genfile('junk.out', 57, 510);
genfile('file1.bin', 57, 100 * 1024);
genfile('file2.out', 42, 99999);

#---------------------------------------------------------------------
runTest('Created check.md5', '-m check.md5', exSuccess,
        <<"END Created check.md5");

Fast File Validator $version by Christopher J. Madsen
Found 2 files

Creating MD5 file check.md5:
1/2 file1.bin
2/2 file2.out
Added 2 files to check.md5
END Created check.md5

#---------------------------------------------------------------------
eq_or_diff(file('check.md5'),
           <<'END Contents of check.md5', 'Contents of check.md5');
MD5(file1.bin) = 8b9330c4cb7fc5b6c4f5af8f8ef9512c
MD5(file2.out) = 917fc62624ecb8b407c8ca693635d99c
END Contents of check.md5

#---------------------------------------------------------------------
runTest('Verified check.md5', '', exSuccess, <<"END Verified check.md5");

Fast File Validator $version by Christopher J. Madsen

check.md5:
1/2 file1.bin ............................................................. OK
2/2 file2.out ............................................................. OK
2 files checked, all ok!
END Verified check.md5

#---------------------------------------------------------------------
genfile('file2.out', 42, 100000);

runTest('Changed file2.out', '', exMismatch, <<"END Changed file2.out");

Fast File Validator $version by Christopher J. Madsen

check.md5:
1/2 file1.bin ............................................................. OK
2/2 file2.out .............................................................BAD
2 files checked
1 file different
END Changed file2.out

#---------------------------------------------------------------------
runTest('Created check.sfv', '-s check.sfv file*', exSuccess,
        <<"END Created check.sfv");

Fast File Validator $version by Christopher J. Madsen
Found 2 files

Creating SFV file check.sfv:
1/2 file1.bin
2/2 file2.out
Added 2 files to check.sfv
END Created check.sfv

#---------------------------------------------------------------------
eq_or_diff(file('check.sfv'),
           <<"END Contents of check.sfv", 'Contents of check.sfv');
; Generated by Fast File Validator $version
;
; Compatibility hack for those poor misguided souls using WIN-SFV32:
; Generated by WIN-SFV32 v1
;
file1.bin FDD6BF96
file2.out 69779FF5
END Contents of check.sfv

#---------------------------------------------------------------------
runTest('Verified check.sfv', 'check.sfv', exSuccess,
        <<"END Verified check.sfv");

Fast File Validator $version by Christopher J. Madsen

check.sfv:
1/2 file1.bin ............................................................. OK
2/2 file2.out ............................................................. OK
2 files checked, all ok!
END Verified check.sfv

#---------------------------------------------------------------------
runTest('Verified test1.md5', "$dataDir/test1.md5", exSuccess,
        <<"END Verified test1.md5");

Fast File Validator $version by Christopher J. Madsen

$dataDir/test1.md5:
1/2 file1.bin ............................................................. OK
2/2 file2.out ............................................................. OK
2 files checked, all ok!
END Verified test1.md5

#---------------------------------------------------------------------
runTest('Verified test1.sfv', "$dataDir/test1.sfv", exSuccess,
        <<"END Verified test1.sfv");

Fast File Validator $version by Christopher J. Madsen

$dataDir/test1.sfv:
1/2 file1.bin ............................................................. OK
2/2 file2.out ............................................................. OK
2 files checked, all ok!
END Verified test1.sfv

#=====================================================================
mkdir 'foo';
mkdir 'foo/bar';
mkdir 'baz';

genfile('foo/alpha.bin',       77,  8472);
genfile('foo/beta.pnf',        94, 32764);
genfile('foo/gamma.bin',       27, 98345);
genfile('foo/bar/delta.bin',   97,  3456);
genfile('foo/bar/epsilon.bin', 97,  1234);
genfile('baz/zeta.bin',        64,  5189);
genfile('baz/eta.bin',        127, 88777);
genfile('baz/theta.bin',      193, 19876);

#---------------------------------------------------------------------
runTest('Created subdirs.md5', '-m subdirs.md5 *.bin *.out foo baz',
        exSuccess, <<"END Created subdirs.md5");

Fast File Validator $version by Christopher J. Madsen
Found 10 files

Creating MD5 file subdirs.md5:
 1/10 baz/eta.bin
 2/10 baz/theta.bin
 3/10 baz/zeta.bin
 4/10 file1.bin
 5/10 file2.out
 6/10 foo/alpha.bin
 7/10 foo/bar/delta.bin
 8/10 foo/bar/epsilon.bin
 9/10 foo/beta.pnf
10/10 foo/gamma.bin
Added 10 files to subdirs.md5
END Created subdirs.md5

#---------------------------------------------------------------------
eq_or_diff(fs(file('subdirs.md5')),
           <<'END Contents of subdirs.md5', 'Contents of subdirs.md5');
MD5(baz/eta.bin) = 3c9d890b5cabc59fe0d13b4c696ad6bd
MD5(baz/theta.bin) = 6adc5c5362b838228ad9c539125606be
MD5(baz/zeta.bin) = c6b36c964e667f8756882374d3f38515
MD5(file1.bin) = 8b9330c4cb7fc5b6c4f5af8f8ef9512c
MD5(file2.out) = bda8c50024a2fb4ac8467b21d719ad0a
MD5(foo/alpha.bin) = dbb831bc5551ac3b68bc47aed2fe8138
MD5(foo/bar/delta.bin) = 03394b0dcf163b15290438750958aa49
MD5(foo/bar/epsilon.bin) = d2b1113d7c67bd82076880c79076efba
MD5(foo/beta.pnf) = c284495f7a9734fd116eb34b5e60a28e
MD5(foo/gamma.bin) = 9fc5b27fc87f63782909a41e1bfd7c20
END Contents of subdirs.md5

#---------------------------------------------------------------------
chdir 'foo';

runTest('Created foo.md5', '-m foo.md5', exSuccess,
        <<"END Created foo.md5");

Fast File Validator $version by Christopher J. Madsen
Found 5 files

Creating MD5 file foo.md5:
1/5 alpha.bin
2/5 bar/delta.bin
3/5 bar/epsilon.bin
4/5 beta.pnf
5/5 gamma.bin
Added 5 files to foo.md5
END Created foo.md5

#---------------------------------------------------------------------
chdir '../baz';

runTest('Created baz.md5', '-m baz.md5', exSuccess,
        <<"END Created baz.md5");

Fast File Validator $version by Christopher J. Madsen
Found 3 files

Creating MD5 file baz.md5:
1/3 eta.bin
2/3 theta.bin
3/3 zeta.bin
Added 3 files to baz.md5
END Created baz.md5

#---------------------------------------------------------------------
chdir '..';

my $expected = <<"END Verified non-rel";

Fast File Validator $version by Christopher J. Madsen

[1/3] subdirs.md5:
 1/10 baz/eta.bin ......................................................... OK
 2/10 baz/theta.bin ....................................................... OK
 3/10 baz/zeta.bin ........................................................ OK
 4/10 file1.bin ........................................................... OK
 5/10 file2.out ........................................................... OK
 6/10 foo/alpha.bin ....................................................... OK
 7/10 foo/bar/delta.bin ................................................... OK
 8/10 foo/bar/epsilon.bin ................................................. OK
 9/10 foo/beta.pnf ........................................................ OK
10/10 foo/gamma.bin ....................................................... OK
10 files checked, all ok!

[2/3] foo/foo.md5:
1/5 alpha.bin .............................................................MISS
2/5 bar/delta.bin .........................................................MISS
3/5 bar/epsilon.bin .......................................................MISS
4/5 beta.pnf ..............................................................MISS
5/5 gamma.bin .............................................................MISS
5 files checked
5 files missing

[3/3] baz/baz.md5:
1/3 eta.bin ...............................................................MISS
2/3 theta.bin .............................................................MISS
3/3 zeta.bin ..............................................................MISS
3 files checked
3 files missing

18 files checked in 3 SFV files, found errors in 2 SFV files
 8 files missing
END Verified non-rel

runTest('Verified non-rel', 'subdirs.md5 foo/foo.md5 baz/baz.md5',
        exMissing, $expected);

runTest('Verified non-rel dirs', 'subdirs.md5 foo baz/',
        exMissing, $expected);

#---------------------------------------------------------------------
$expected = <<"END Verified --relative";

Fast File Validator $version by Christopher J. Madsen

[1/3] subdirs.md5:
 1/10 baz/eta.bin ......................................................... OK
 2/10 baz/theta.bin ....................................................... OK
 3/10 baz/zeta.bin ........................................................ OK
 4/10 file1.bin ........................................................... OK
 5/10 file2.out ........................................................... OK
 6/10 foo/alpha.bin ....................................................... OK
 7/10 foo/bar/delta.bin ................................................... OK
 8/10 foo/bar/epsilon.bin ................................................. OK
 9/10 foo/beta.pnf ........................................................ OK
10/10 foo/gamma.bin ....................................................... OK
10 files checked, all ok!

[2/3] foo/foo.md5:
1/5 foo/alpha.bin ......................................................... OK
2/5 foo/bar/delta.bin ..................................................... OK
3/5 foo/bar/epsilon.bin ................................................... OK
4/5 foo/beta.pnf .......................................................... OK
5/5 foo/gamma.bin ......................................................... OK
5 files checked, all ok!

[3/3] baz/baz.md5:
1/3 baz/eta.bin ........................................................... OK
2/3 baz/theta.bin ......................................................... OK
3/3 baz/zeta.bin .......................................................... OK
3 files checked, all ok!

18 files checked in 3 SFV files, all ok!
END Verified --relative

runTest('Verified --relative', '-r subdirs.md5 foo/foo.md5 baz/baz.md5',
        exSuccess, $expected);

runTest('Verified --relative dirs', '-r subdirs.md5 foo/ baz',
        exSuccess, $expected);

#---------------------------------------------------------------------
my $usage = <<"END --no-such-option";
--no-such-option is not a recognized option

Fast File Validator $version by Christopher J. Madsen

Usage: FFV [options] [FILE ...]
  -a, --append          Append to existing checksum file (when creating one)
  -o, --overwrite       Replace an existing checksum file (when creating one)
  -e, --allow-errors    Don't stop if there's an error reading a file
  -h, --hidden          Include hidden files when creating a checksum file
  -m, --create-md5=FILE Create an MD5 checksum file
  -s, --create-sfv=FILE Create a SFV (CRC-32) file
  -n, --no-verify       Just check if files exist (don't verify checksums)
  -p, --pause           Wait for keypress before exiting
  -r, --relative        Interpret paths relative to the checksum file
] -w, --allow-writing   Process files that are being written by other programs
  -?, --help            Display this help message
      --license         Display license information
      --version         Display version information

Note: You can't combine the --create-sfv and --create-md5 options
      (or the --append and --overwrite options, either).
END --no-such-option

if ($Win32) {
  $usage =~ s/^\]/ /mg;
} else {
  $usage =~ s/^\].+\n//mg;      # Remove Win32-only options
}

runTest('--no-such-option', '--no-such-option', exUsage, $usage);

#---------------------------------------------------------------------
chdir '..';   # Otherwise Windows can't remove the temporary directory
