#! /usr/bin/perl
#---------------------------------------------------------------------
# $Id: zipdist.pl 4827 2011-11-20 23:30:46Z cjm $
# Copyright 2005 Christopher J. Madsen
#
# Generate the Win32 distribution ZIP file
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of
# the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#---------------------------------------------------------------------

use strict;
use FindBin '$Bin';
use File::Spec ();
use IO::All;

$| = 1;                         # Disable buffering

my @unlink;

#---------------------------------------------------------------------
sub getFrom
{
  my ($from, $to) = @_;

  return if -e $to and not -e $from;

  Win32::CopyFile($from, $to, 0) or die "Can't copy $from to $to\n";
  push @unlink, $to;
} # end getFrom

#---------------------------------------------------------------------
sub makeZip
{
  my $fn   = shift @_;
  my $opts = shift @_;

  die "$fn exists" if -e $fn;

  print "Creating $fn...\n";

  open(ZIP, "| zip -9X\@$opts $fn") or die;

  foreach (@_) {
    die "$_ not found" unless -e $_;
    print ZIP "$_\n"
  }

  close ZIP or die;
} # end makeZip

#---------------------------------------------------------------------
# Get version from configure.ac:

our ($name, $major, $minor);
do "$Bin/getversion.pm";

my $zip = sprintf "FFV_%d%02d.zip", $major, $minor;
die "$zip exists\n" if -e "$Bin/../$zip";
die "Source.zip exists\n" if -e "$Bin/../Source.zip";

my $exe = File::Spec->rel2abs("../win32/Release/FFV.exe", $Bin);
die "Need to rebuild Release/FFV.exe\n" unless -e $exe and -M _ < 10/(60*24);

# Build rndfile.exe:

chdir $Bin or die "Can't cd to $Bin: $!";
open(IN, '<', 'rndfile.c') or die "Can't open rndfile.c: $!";

my $command;
while (<IN>) {
  $command = $1, last if m!^/\*\s+compile-command:\s*"(.+)"\s*\*/$!;
}
close IN;
die "Can't find compile-command in rndfile.c" unless defined $command;

system $command and die "Compiling rndfile.c failed\n";

# Run the tests:

chdir "$Bin/.." or die "Can't cd to $Bin/..: $!";

$ENV{FFV_EXE_PATH} = $exe;
print "\n";
system prove => 'tools/testFFV.pl' and die "Tests failed\n";
print "\n";

# Read Makefile.am:

my $conf = io('Makefile.am')->scalar;

$conf =~ s/ +\\\n[ \t]*/ /g;    # Merge up continued lines

# Parse out the source files:

my @testFiles = grep { not /[~#]$/ }
    ('tools/rndfile.exe', <tools/test*.pl>, <tools/testdata/*>);
my %exclude = map { $_ => 1 } @testFiles;

my @files = qw(Makefile.am config.h.in configure.ac tools/rndfile.c);
while ($conf =~ /^\w+_(?:SOURCES|DIST) +=(.+)/mg) {
  push @files, grep { not $exclude{$_} } split(' ', $1);
}

# Zip up the source files:

makeZip('Source.zip', '', @files);

# Generate documentation:

system qw(perl tools/genfile.pl -D Win32 tools/FFV.pod.tt tools/ReadMe.tt);

getFrom('AUTHORS', 'AUTHORS.txt');
getFrom('\emacs\21.3\etc\COPYING', 'COPYING.txt');

# Create ZIP file:

getFrom('win32/Release/FFV.exe', 'FFV.exe');

makeZip($zip, '', qw(ReadMe.txt AUTHORS.txt COPYING.txt
                     FFV.exe FFV.txt), @testFiles,
                  qw(Source.zip));

# Clean up:

unlink 'Source.zip';

END { unlink @unlink; }
